/*******************  start of original comments  ********************/
/*
 * Written by Douglas Thomson (1989/1990)
 *
 * This source code is released into the public domain.
 */

/*
 * Name:    dte - Doug's Text Editor program - main editor module
 * Purpose: This file contains the main editor module, and a number of the
 *           smaller miscellaneous editing commands.
 *          It also contains the code for dispatching commands.
 * File:    ed.c
 * Author:  Douglas Thomson
 * System:  this file is intended to be system-independent
 * Date:    October 1, 1989
 * I/O:     file being edited
 *          files read or written
 *          user commands and prompts
 * Notes:   see the file "dte.doc" for general program documentation
 */
/*********************  end of original comments   ********************/

/*
 * The basic editor routines have been EXTENSIVELY reworked.  I have added
 * support for lines longer than 80 columns and I have added line number
 * support.  I like to know the real line number that editor functions are
 * working on and I like to know the total number of lines in a file.
 *
 * I rewrote the big series of ifs in the dispatch subroutine.  It is now
 * an array of pointers to functions.  We know what function to call as soon
 * as a key is pressed.  It is also makes it easier to implement a configuration
 * utility.
 *
 * I added a few functions that I use quite often and I deleted a few that I
 * rarely use.  Added are Split Line, Join Line, and Duplicate Line.   Deleted
 * are Goto Marker 0-9 (others?).
 *
 * ************ In tde 1.3, I put Goto Marker 0-9 back in.  ***************
 *
 * I felt that the insert routine should be separated into two routines.  One
 * for inserting the various combinations of newlines and one for inserting
 * 'regular' text characters, ASCII and extended ASCII characters.
 *
 * One of Doug's design considerations was keeping screen updates to a minimum.
 * I have expanded upon that idea and added support for updating windows
 * LOCALly, GLOBALly, or NOT_LOCALly.  For example, scrolling in one window
 * does not affect the text in another window - LOCAL update.  Adding, deleting,
 * or modifying text in one window may affect text in other windows - GLOBAL
 * update.  Sometimes, updates to the current window are handled in the task
 * routines so updates to other windows are done NOT_LOCALly.
 *
 * Also note that using functions copy_line and un_copy_line to change a line
 * automatically adjusts the g_status.end_mem pointer.  If a function bypasses
 * those functions, adjusting the g_status.end_mem pointer must be done
 * explicitly.
 *
 * New editor name:  tde, the Thomson-Davis Editor.
 * Author:           Frank Davis
 * Date:             June 5, 1991, version 1.0
 * Date:             July 29, 1991, version 1.1
 * Date:             October 5, 1991, version 1.2
 * Date:             January 20, 1992, version 1.3
 *
 * This modification of Douglas Thomson's code is released into the
 * public domain, Frank Davis.   You may distribute it freely.
 */

#include "tdestr.h"     /* global variables */
#include "define.h"
#include "tdefunc.h"
#include "global.h"     /* global variables */


/*
 * Name:    tab_key
 * Purpose: To make the necessary changes after the user types the tab key.
 * Date:    June 5, 1991
 * Passed:  window: information allowing access to the current window
 * Notes:   If in insert mode, then this function adds the required
 *           number of spaces in the file.
 *          If not in insert mode, then tab simply moves the cursor right
 *           the required distance.
 */
void tab_key( windows *window )
{
int spaces;      /* the spaces to move to the next tab stop */
char *source;    /* source for block move to make room for c */
char *dest;      /* destination for block move */
int pad, len;
register int rcol;
register windows *win;  /* put window pointer in a register */

   win  = window;
   if (*win->cursor == CONTROL_Z)
      return;
   rcol = win->rcol;
   /*
    * work out the number of spaces to the next tab stop
    */
   spaces = mode.tab_size - (rcol % mode.tab_size);

   if (mode.insert && rcol + spaces < g_display.line_length) {
      copy_line( win->cursor, win->bottom_line );
      /*
       * work out how many characters need to be inserted
       */
      len = linelen( g_status.line_buff );
      pad = rcol > len ? rcol - len : 0;
      if (g_status.line_buff[len] == CONTROL_Z)
         ++pad;
      if (len + pad + spaces >= g_display.line_length)
         error( WARNING, win->bottom_line, "line too long to add" );
      else {
         if (pad > 0  || spaces > 0) {
            if (g_status.line_buff[len] == CONTROL_Z) {
               g_status.line_buff[len] = '\n';
               g_status.line_buff[len+1] = CONTROL_Z;
               ++win->file_info->length;
               show_size( window );
               --pad;
               ++len;
            }
            source = g_status.line_buff + rcol - pad;
            dest = source + pad + spaces;
            memmove( dest, source, len+pad-rcol+2 );

            /*
             * if padding was required, then put in the required spaces
             */
            memset( source, ' ', pad + spaces );
         }

         win->file_info->dirty = GLOBAL;
         show_changed_line( win );
         rcol += spaces;
         win->ccol += spaces;
      }
   } else if (rcol + spaces <= g_display.line_length) {
      /*
       * advance the cursor without changing the text underneath
       */
      rcol += spaces;
      win->ccol += spaces;
   }
   check_virtual_col( win, rcol, win->ccol );
}


/*
 * Name:    backtab
 * Purpose: To make the necessary changes after the user presses the backtab.
 * Date:    November 1, 1991
 * Passed:  window: information allowing access to the current window
 * Notes:   If in insert mode, then this function subs the required
 *           number of spaces in the file.
 *          If not in insert mode, then tab simply moves the cursor left
 *           the required distance.
 */
void backtab( windows *window )
{
int spaces;      /* the spaces to move to the next tab stop */
char *source;    /* source for block move to make room for c */
char *dest;      /* destination for block move */
int pad, len;
register int rcol;
register windows *win;  /* put window pointer in a register */

   win  = window;
   rcol = win->rcol;
   if (*win->cursor == CONTROL_Z || win->rcol == 0)
      return;
   /*
    * work out the number of spaces to the previous tab stop
    */
   spaces = win->rcol % mode.tab_size;
   if (spaces == 0)
      spaces = mode.tab_size;
   copy_line( win->cursor, win->bottom_line );
   len = linelen( g_status.line_buff );
   if (mode.insert && rcol - spaces < len) {
      pad = rcol > len ? rcol - len : 0;
      if (g_status.line_buff[len] == CONTROL_Z)
         ++pad;
      if (pad > 0  || spaces > 0) {
         if (g_status.line_buff[len] == CONTROL_Z) {
            g_status.line_buff[len] = '\n';
            g_status.line_buff[len+1] = CONTROL_Z;
            ++win->file_info->length;
            show_size( win );
            --pad;
            ++len;
         }
         /*
          * if padding was required, then put in the required spaces
          */
         if (pad > 0) {
            source = g_status.line_buff + rcol - pad;
            dest = source + pad;
            memmove( dest, source, pad+2 );
            memset( source, ' ', pad );
         }
         source = g_status.line_buff + rcol;
         dest = source - spaces;
         memmove( dest, source, len+pad-rcol+2 );
      }

      win->file_info->dirty = GLOBAL;
      show_changed_line( win );
      rcol -= spaces;
      win->ccol -= spaces;
   } else {
      /*
       * move the cursor without changing the text underneath
       */
      rcol -= spaces;
      if (rcol < 0)
         rcol = 0;
      win->ccol -= spaces;
   }
   check_virtual_col( win, rcol, win->ccol );
}


/*
 * Name:    insert_newline
 * Purpose: insert a newline
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   There a several ways to insert a line into a file:  1) pressing
 *          a key, 2) word wrap, 3) any others?
 *          When doing word wrap or format paragraph, don't show any changes.
 *            Wait until the function finishes then show all changes at once.
 */
void insert_newline( windows *window )
{
char *source;       /* source for block move to make room for c */
char *dest;         /* destination for block move */
int len;            /* length of current line */
register int add;   /* characters to be added (usually 1 in insert mode) */
int rcol;
text_ptr prev;      /* previous lines scanned for autoindent */
long length;
int carriage_return;
int split_line;
int wordwrap;
int dirty;
register windows *win;  /* put window pointer in a register */

   win = window;
   length = win->file_info->length;
   if (win->rline > length && *win->cursor != CONTROL_Z)
      return;
   wordwrap = mode.word_wrap;
   switch (g_status.command) {
      case WordWrap:
         wordwrap = TRUE;
      case Rturn :
         carriage_return = TRUE;
         split_line = FALSE;
         break;
      case AddLine :
         split_line = carriage_return = FALSE;
         break;
      case SplitLine :
         split_line = carriage_return = TRUE;
         break;
   }

   /*
    * make window temporarily invisible to the un_copy_line function
    */
   win->visible = FALSE;
   win->cursor = cpf( win->cursor );
   copy_line( win->cursor, win->bottom_line );
   len = linelen( g_status.line_buff );

   source = g_status.line_buff + len;
   if (carriage_return || split_line) {
      if (win->rcol < len)
         source = g_status.line_buff + win->rcol;
   }
   /*
    *  make room for '\n' just after source (source+1)
    */
   memmove( source+1, source, linelen( source )+2 );

   *source = '\n';
   un_copy_line( win->cursor, win, TRUE );
   adjust_windows_cursor( win, 1 );

   ++win->file_info->length;
   win->file_info->dirty = NOT_LOCAL;
   if (length == 0l || wordwrap || win->cline == win->bottom_line)
      win->file_info->dirty = GLOBAL;
   else if (!split_line)
      update_line( win );

   /*
    * If the cursor is to move down to the next line, then update
    *  the line and column appropriately.
    */
   if (carriage_return || split_line) {
      dirty = win->file_info->dirty;
      win->cursor = find_next( win->cursor );
      if (win->cline < win->bottom_line)
         win->cline++;
      win->rline++;
      rcol = win->rcol;

      /*
       * indentation is only required if we are in the right mode,
       *  the user typed <CR>, and if there is not space followed
       *  by something after the cursor.
       */
      if (mode.indent || wordwrap) {
         /*
          * autoindentation is required. Match the indentation of
          *  the first line above that is not blank.
          */
         if (wordwrap) {
            prev = cpb( win->cursor );
            prev = find_prev( prev );
            if (prev == NULL)
               add = mode.left_margin;
            else if (linelen( prev ) == 0)
               add = mode.parg_margin;
            else
               add = mode.left_margin;
         } else {
            add = first_non_blank( g_status.line_buff );
            if (g_status.line_buff[add] == '\n' ||
                      g_status.line_buff[add] == CONTROL_Z) {
               prev = cpb( win->cursor );
               while ((prev = find_prev( prev )) != NULL) {
                  add = first_non_blank( prev );
                  if (prev[add] != '\n')
                     break;
               }
            }
         }
         copy_line( win->cursor, win->bottom_line );
         len = linelen( g_status.line_buff );
         source = g_status.line_buff;
         if (len + add > MAX_LINE_LENGTH)
            add = MAX_LINE_LENGTH - len;
         dest = source + add;
         memmove( dest, source, len+2 );

         /*
          * now put in the autoindent characters
          */
         memset( source, ' ', add );
         win->rcol = add;
         un_copy_line( win->cursor, win, TRUE );
      } else
         win->rcol = 0;
      if (split_line) {
         win->cursor = cpb( win->cursor );
         win->cursor = find_prev( win->cursor );
         if (win->cline > win->top_line)
            win->cline--;
         win->rline--;
         win->rcol = rcol;
      }
      check_virtual_col( win, win->rcol, win->ccol );
      if (dirty == GLOBAL || win->file_info->dirty == LOCAL || wordwrap)
         win->file_info->dirty = GLOBAL;
      else
         win->file_info->dirty = dirty;
   }

   /*
    * record that file has been modified
    */
   if (win->file_info->dirty != GLOBAL)
      my_scroll_down( win );
   restore_marked_block( win, 1 );
   show_size( win );
   win->visible = TRUE;
}


/*
 * Name:    insert_overwrite
 * Purpose: To make the necessary changes after the user has typed a normal
 *           printable character
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 */
void insert_overwrite( windows *window )
{
char *source;       /* source for block move to make room for c */
char *dest;         /* destination for block move */
int len;            /* length of current line */
int pad;            /* padding to add if cursor beyond end of line */
int add;            /* characters to be added (usually 1 in insert mode) */
register int rcol;
register windows *win;  /* put window pointer in a register */

   win = window;
   if (*win->cursor == CONTROL_Z || g_status.key_pressed >= 256)
      return;
   rcol = win->rcol;
   /*
    * first check we have room - the editor can not
    *  cope with lines wider than g_display.line_length
    */
   if (rcol >= g_display.line_length)
      error( WARNING, win->bottom_line, "cannot insert more characters" );
   else {
      copy_line( win->cursor, win->bottom_line );

      /*
       * work out how many characters need to be inserted
       */
      len = linelen( g_status.line_buff );
      pad = rcol > len ? rcol - len : 0;

      /*
       * if this is the last line in a file, the last character in the
       * line buffer will be CONTROL_Z.  increment pad and insert a \n.
       */
      if (g_status.line_buff[len] == CONTROL_Z)
         ++pad;

      if (mode.insert || rcol >= len)
         /*
          * inserted characters, or overwritten characters at the end of
          *  the line, are inserted.
          */
         add = 1;
      else
         /*
          *  and no extra space is required to overwrite existing characters
          */
         add = 0;

      /*
       * check that current line would not get too long.
       */
      if (len + pad + add >= g_display.line_length)
         error( WARNING, win->bottom_line, "no more room to add" );
      else {

         /*
          * make room for whatever needs to be inserted
          */
         if (pad > 0  || add > 0) {
            source = g_status.line_buff + len;
            if (*source == CONTROL_Z) {
               if (rcol > len)
                  source = g_status.line_buff + rcol + 1;
               *source++ = '\n';
               *source   = CONTROL_Z;
               ++win->file_info->length;
               show_size( win );
               --pad;
               ++len;
            }
            source = g_status.line_buff + rcol - pad;
            dest = source + pad + add;
            memmove( dest, source, len + pad - rcol + 2 );
            /*
             * put in the required padding
             */
            memset( source, ' ', pad );
         }
         g_status.line_buff[rcol] = (char)g_status.key_pressed;

         /*
          * always increment the real column (rcol) then adjust the
          * logical and base column as needed.   show the changed line
          * in all but the LOCAL window.  In the LOCAL window, there are
          * two cases:  1) update the line, or 2) redraw the window if
          * cursor goes too far right.
          */
         win->file_info->dirty = NOT_LOCAL;
         show_changed_line( win );
         if (win->ccol < g_display.ncols - 1) {
            show_curl_line( win );
            win->ccol++;
         } else {
            win->bcol++;
            win->file_info->dirty = LOCAL;
         }
         rcol++;
      }

      /*
       * record that file has been modified and adjust cursors,
       * file start and end pointers as needed.
       */
      check_virtual_col( win, rcol, win->ccol );
      win->file_info->modified = TRUE;
      if (mode.word_wrap) {
         g_status.command = WordWrap;
         word_wrap( win );
      }
   }
}


/*
 * Name:    join_line
 * Purpose: To join current line and line below at cursor
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   trunc the line then join with line below if it exists
 */
void join_line( windows *window )
{
register int len;   /* length of current line */
char *source;       /* source for block move to delete word */
char *dest;         /* destination for block move */
text_ptr p;         /* next line in file */
int pad, i;         /* padding spaces required */
int cr;             /* does current line end with carriage return? */
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->rline > win->file_info->length || *win->cursor == CONTROL_Z)
      return;

   win->cursor = cpf( win->cursor );
   load_undo_buffer( win->cursor );
   copy_line( win->cursor, win->bottom_line );
   if (win->rcol < (len = linelen( g_status.line_buff ))) {
      /*
       * delete rest of line
       */
      dest = g_status.line_buff + win->rcol;

      if (g_status.line_buff[len] == '\n')
         *dest++ = '\n';

      *dest = CONTROL_Z;
      un_copy_line( win->cursor, win, FALSE );
      win->file_info->dirty = GLOBAL;
   }
   /*
    * we need to combine with the next line, if any
    */
   if ((p = find_next( win->cursor )) != NULL && *p != CONTROL_Z) {
      /*
       * add padding if required
       */
      len = linelen( g_status.line_buff );
      cr  = g_status.line_buff[len] == '\n' ? 1 : 0;
      pad = win->rcol > len ? win->rcol - len : 0;

      /*
       * check room to combine lines
       */
      if (len + pad + cr + linelen( p ) >= g_display.line_length)
         error( WARNING, win->bottom_line, "cannot combine lines" );
      else {

         /*
          * do the move
          */
         source = g_status.line_buff + win->rcol - pad;
         dest   = source + pad;
         memmove( dest, source, len + pad - win->rcol + 1 + cr );

         /*
          * insert the padding
          */
         while (pad--)
            *source++ = ' ';

         /*
          * remove the \n separating the two lines.
          */
         i = 0;
         if (*source == '\n') {
            *source = CONTROL_Z;
            i = -1;
         }
         g_status.copied = TRUE;
         un_copy_line( win->cursor, win, FALSE );
         adjust_windows_cursor( win, i );
         --win->file_info->length;
         restore_marked_block( win, -1 );
         show_size( win );
         win->file_info->dirty = GLOBAL;
      }
   }
}


/*
 * Name:    word_delete
 * Purpose: To delete from the cursor to the start of the next word.
 * Date:    September 1, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   If the cursor is at the right of the line, then combine the
 *           current line with the next one, leaving the cursor where it
 *           is.
 *          If the cursor is on an alphanumeric character, then all
 *           subsequent alphanumeric characters are deleted.
 *          If the cursor is on a space, then all subsequent spaces
 *           are deleted.
 *          If the cursor is on a punctuation character, then all
 *           subsequent punctuation characters are deleted.
 */
void word_delete( windows *window )
{
int len;            /* length of current line */
register int start; /* column that next word starts in */
char *source;       /* source for block move to delete word */
char *dest;         /* destination for block move */
int alpha;          /* is the cursor char alphanumeric? */
text_ptr p;
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->rline > win->file_info->length || *win->cursor == CONTROL_Z)
      return;
   win->cursor = cpf( win->cursor );
   copy_line( win->cursor, win->bottom_line );
   if (win->rcol >= (len = linelen( g_status.line_buff ))) {
      join_line( win );
      p = win->cursor + win->rcol;
      if (*p != CONTROL_Z)
         load_undo_buffer( p );
   } else {
      /*
       * normal word delete
       *
       * find the start of the next word
       */
      start = win->rcol;
      if (g_status.line_buff[start] == ' ') {
         /*
          * the cursor was on a space, so eat all consecutive spaces
          *  from the cursor onwards.
          */
         while (g_status.line_buff[start] == ' ')
            ++start;
      } else {
         /*
          * eat all consecutive characters in the same class (spaces
          *  are considered to be in the same class as the cursor
          *  character)
          */
         alpha = myisalnum( g_status.line_buff[start++] );
         while (start < len) {
            if (g_status.line_buff[start] == ' ')
                /*
                 * the next character that is not a space will
                 *  end the delete
                 */
                alpha = -1;
            else if (alpha != myisalnum( g_status.line_buff[start] )) {
                if (g_status.line_buff[start] != ' ')
                    break;
            }
            ++start;
         }
      }

      /*
       * move text to delete word
       */
      source = g_status.line_buff + start;
      dest = g_status.line_buff + win->rcol;
      memmove( dest, source, len-start+2 );
      win->file_info->modified = TRUE;
      win->file_info->dirty = GLOBAL;
      if (g_status.command == WordDelete)
         show_changed_line( win );
   }
}


/*
 * Name:    dup_line
 * Purpose: Duplicate current line
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   cursor stays on current line
 */
void dup_line( windows *window )
{
register int len;       /* length of current line */
long number;
text_ptr d, s;
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->rline > win->file_info->length)
      return;
   win->cursor = cpf( win->cursor );

   un_copy_line( win->cursor, win, TRUE );
   /*
    * don't dup the ^Z or a NULL line
    */
   if (*win->cursor != CONTROL_Z && (d=find_next( win->cursor )) != NULL) {

      /*
       * don't use buffers to dup the line.  use hw_move to make space and
       * copy current line at same time.  d is set to beginning of next line.
       */
      s = win->cursor;
      len = linelen( s );
      if (s[len] == '\n')
         ++len;
      number = ptoul( g_status.end_mem ) - ptoul( s );
      hw_move( d, s, number );
      ++win->file_info->length;
      g_status.end_mem = addltop( len, g_status.end_mem );
      adjust_start_end( win->file_info, len );
      addorsub_all_cursors( win, len );
      adjust_windows_cursor( win, 1 );

      /*
       * if current line is the bottom line, we can't see the dup line because
       * cursor doesn't move and dup line is added after current line.
       */
      if  (win->cline != win->bottom_line)
         my_scroll_down( win );
      win->file_info->dirty = NOT_LOCAL;

      /*
       * record that file has been modified
       */
      win->file_info->modified = TRUE;
      restore_marked_block( win, 1 );
      show_size( win );
      show_avail_mem( );
   } else
      error( WARNING, win->bottom_line, "cannot duplicate line" );
}


/*
 * Name:    back_space
 * Purpose: To delete the character to the left of the cursor.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   If the cursor is at the left of the line, then combine the
 *           current line with the previous one.
 *          If in indent mode, and the cursor is on the first non-blank
 *           character of the line, then match the indentation of an
 *           earlier line.
 */
void back_space( windows *window )
{
int len;            /* length of the current line */
char *source;       /* source of block move to delete character */
char *dest;         /* destination of block move */
text_ptr p;         /* previous line in file */
int plen;           /* length of previous line */
int del_count;      /* number of characters to delete */
int pos;            /* the position of the first non-blank char */
register int rcol;
int ccol;
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->rline > win->file_info->length || *win->cursor == CONTROL_Z)
      return;
   win->cursor = cpf( win->cursor );
   copy_line( win->cursor, win->bottom_line );
   len = linelen( g_status.line_buff );
   rcol = win->rcol;
   ccol = win->ccol;
   if (rcol == 0) {
      /*
       * combine this line with the previous, if any
       */
      win->cursor = cpb( win->cursor );
      if ((p = find_prev( win->cursor )) != NULL) {
         if (len + 2 + (plen = linelen( p )) >= g_display.line_length) {
            error( WARNING, win->bottom_line, "cannot combine lines" );
            return;
         }

         un_copy_line( win->cursor, win, TRUE );
         copy_line( p, win->bottom_line );
         load_undo_buffer( p );
         g_status.line_buff[plen] = CONTROL_Z;
         win->cursor = p;
         un_copy_line( win->cursor, win, FALSE );

         /*
          * make sure cursor stays on the screen, at the end of the
          *  previous line
          */
         if (win->cline > win->top_line)
            --win->cline;
         --win->rline;
         rcol = plen;
         ccol = rcol - win->bcol;
         --win->file_info->length;
         restore_marked_block( win, -1 );
         adjust_windows_cursor( win, -1 );
         show_size( win );
         check_virtual_col( win, rcol, ccol );
         win->file_info->dirty = GLOBAL;
      }
   } else {
      /*
       * normal delete
       *
       * find out how much to delete (depends on indent mode)
       */
      del_count = 1;   /* the default */
      if (mode.indent) {
         /*
          * indent only happens if the cursor is on the first
          *  non-blank character of the line
          */
         if ((pos = first_non_blank( g_status.line_buff )) == rcol
                    || g_status.line_buff[pos] == '\n'
                    || g_status.line_buff[pos] == CONTROL_Z) {
            /*
             * now work out how much to indent
             */
            p = cpb( win->cursor );
            for (p=find_prev( p ); p != NULL; p=find_prev( p )) {
               if ((plen=first_non_blank( p )) < rcol && *(p+plen)!='\n') {
                  /*
                   * found the line to match
                   */
                  del_count = rcol - plen;
                  break;
               }
            }
         }
      }

      /*
       * move text to delete char(s), unless no chars actually there
       */
      if (rcol - del_count < len) {
         dest = g_status.line_buff + rcol - del_count;
         if (rcol > len) {
            source = g_status.line_buff + len;
            len = 2;
         } else {
            source = g_status.line_buff + rcol;
            len = len - rcol + 2;
         }
         memmove( dest, source, len );
      }
      rcol -= del_count;
      ccol -= del_count;
      check_virtual_col( win, rcol, ccol );
      win->file_info->dirty = GLOBAL;
      show_changed_line( win );
   }
   win->file_info->modified = TRUE;
}


/*
 * Name:    line_kill
 * Purpose: To delete the line the cursor is on.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   If *window->cursor is pointing to CONTROL_Z then do not do a
 *          line kill (can't kill a NULL line).
 */
void line_kill( windows *window )
{
int i = 0;
text_ptr s;         /* next line in file */
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->file_info->length > 0  && *win->cursor != CONTROL_Z) {
      load_undo_buffer( g_status.copied ? g_status.line_buff : win->cursor);
      g_status.copied = TRUE;
      g_status.line_buff[0] = CONTROL_Z;
      s = win->cursor = cpf( win->cursor );
      /*
       * if line to delete has \n at end of line then decrement file length.
       */
      if (*(s + linelen( s )) == '\n') {
         --win->file_info->length;
         --i;
      }
      un_copy_line( s, win, FALSE );
      win->file_info->dirty = NOT_LOCAL;

      /*
       * move all cursors one according to i, restore begin and end block
       */
      adjust_windows_cursor( win, i );
      restore_marked_block( win, i );

      /*
       * we are not doing a GLOBAL update, so update current window here
       */
      if (win->file_info->dirty == NOT_LOCAL)
         my_scroll_down( win );
      show_size( win );
   }
}


/*
 * Name:    char_del_under
 * Purpose: To delete the character under the cursor.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   If the cursor is beyond the end of the line, then this
 *           command is ignored.
 *          DeleteChar and StreamDeleteChar use this function.
 */
void char_del_under( windows *window )
{
char *source;    /* source of block move to delete character */
register int len;
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->rline > win->file_info->length || *win->cursor == CONTROL_Z)
      return;
   copy_line( win->cursor, win->bottom_line );
   if (win->rcol < (len = linelen( g_status.line_buff ))) {
      /*
       * move text to delete using buffer
       */
      source = g_status.line_buff + win->rcol + 1;
      memmove( source-1, source, len - win->rcol + 2 );
      win->file_info->dirty    = GLOBAL;
      win->file_info->modified = TRUE;
      show_changed_line( win );
   } else if (g_status.command == StreamDeleteChar)
      join_line( win );
}


/*
 * Name:    eol_kill
 * Purpose: To delete everything from the cursor to the end of the line.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   If the cursor is beyond the end of the line, then this
 *           command is ignored.
 */
void eol_kill( windows *window )
{
register char *dest;  /* the start of the delete area */
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->rline > win->file_info->length || *win->cursor == CONTROL_Z)
      return;
   copy_line( win->cursor, win->bottom_line );
   load_undo_buffer( g_status.line_buff );
   if (win->rcol < linelen( g_status.line_buff )) {
      /*
       * truncate to delete rest of line
       */
      dest = g_status.line_buff + win->rcol;
      *dest++ = '\n';
      *dest = CONTROL_Z;
      win->file_info->dirty = GLOBAL;
      show_changed_line( win );
   }
}


/*
 * Name:    undo_line
 * Purpose: To retrieve unaltered line if possible.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   Changes are made to the line buffer so the underlying text has
 *          not changed.  Put the unchanged line from the file into the
 *          line buffer and display it.
 */
void undo_line( windows *window )
{
register windows *win;  /* put window pointer in a register */

   win = window;
   if (win->rline > win->file_info->length)
      return;
   if (g_status.copied) {
      g_status.copied = FALSE;
      copy_line( win->cursor, win->bottom_line );
      win->file_info->dirty = GLOBAL;
      show_changed_line( win );
   }
}


/*
 * Name:    undo
 * Purpose: To retrieve (pop) a line from the undo stack
 * Date:    September 26, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   Insert an empty line into the file then pop the line in the undo
 *          stack.  When we pop line 0, there are no more lines on the stack.
 *          Set the stack pointer to -1 to indicate an empty stack.
 */
void undo( windows *window )
{
char *source;
char *undo_line;
int len;
int head;
register windows *win;  /* put window pointer in a register */

   if (g_status.undo_head < 0)
      return;
   --g_status.undo_head;
   win = window;
   win->cursor = cpf( win->cursor );
   un_copy_line( win->cursor, win, TRUE );
   copy_line( win->cursor, win->bottom_line );

   source = g_status.line_buff;
   /*
    *  make room for '\n'.  then, after we un_copy the g_status.line_buff, we
    *  have added a line to the file.
    */
   memmove( source+1, source, linelen( source )+2 );
   *source = '\n';
   un_copy_line( win->cursor, win, TRUE );

   /*
    *  ajust cursors in other windows opened to the same file.
    */
   adjust_windows_cursor( win, 1 );

   /*
    * copy the line in the top of the stack to the line_buffer.  then, copy
    * the line buffer to the file.
    */
   g_status.copied = TRUE;
   head = g_status.undo_head;
   undo_line = &g_status.undo_buffer[head][0];
   len = find_CONTROL_Z( undo_line ) + 1;
   memcpy( source, undo_line, len );
   un_copy_line( win->cursor, win, TRUE );

   /*
    * we have now undeleted a line.  increment the file length and display
    * it.
    */
   win->file_info->length++;
   win->file_info->dirty = GLOBAL;
   show_size( win );

   /*
    * if g_status.undo_head == 0, we just popped the last line in the stack.
    * set the stack pointer to -1 to show there are no more lines.
    */
   if (g_status.undo_head == 0)
      g_status.undo_head = -1;
}


/*
 * Name:    beg_next_line
 * Purpose: To move the cursor to the beginning of the next line.
 * Date:    October 4, 1991
 * Passed:  window:   information allowing access to the current window
 */
void beg_next_line( windows *window )
{
   window->rcol = 0;
   check_virtual_col( window, window->rcol, window->ccol );
   move_down( window );
}


/*
 * Name:    next_line
 * Purpose: To move the cursor to the first character of the next line.
 * Date:    October 4, 1991
 * Passed:  window:   information allowing access to the current window
 */
void next_line( windows *window )
{
register int rcol;
register windows *win;  /* put window pointer in a register */

   win = window;
   move_down( win );
   rcol = first_non_blank( win->cursor );
   if (win->cursor[rcol] == '\n')
      rcol = 0;
   check_virtual_col( win, rcol, win->ccol );
}


/*
 * Name:    home
 * Purpose: To move the cursor to the left of the current line.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 */
void home( windows *window )
{
register int rcol;
register windows *win;  /* put window pointer in a register */

   win = window;
   if (g_status.copied) {
      rcol = first_non_blank( g_status.line_buff );
      if (g_status.line_buff[rcol] == '\n')
         rcol = 0;
   } else {
      win->cursor = cpf( win->cursor );
      rcol = first_non_blank( win->cursor );
      if (win->cursor[rcol] == '\n')
         rcol = 0;
   }
   if (win->rcol == rcol)
      rcol = 0;
   check_virtual_col( win, rcol, win->ccol );
   sync( win );
}


/*
 * Name:    goto_right
 * Purpose: To move the cursor to the right of the current line.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 */
void goto_right( windows *window )
{
register int rcol;
register windows *win;  /* put window pointer in a register */

   win = window;
   rcol = linelen( g_status.copied ? g_status.line_buff : win->cursor );
   win->ccol = rcol - win->bcol;
   check_virtual_col( win, rcol, win->ccol );
   sync( win );
}


/*
 * Name:    goto_top
 * Purpose: To move the cursor to the top of the current window.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   If the start of the file occurs before the top of the window,
 *           then the start of the file is moved to the top of the window.
 */
void goto_top( windows *window )
{
text_ptr cursor;  /* anticipated cursor line */
register windows *win;  /* put window pointer in a register */

   win = window;
   un_copy_line( win->cursor, win, TRUE );
   update_line( win );
   win->cursor = cpb( win->cursor );
   for (; win->cline > win->top_line; win->cline--,win->rline--) {
      if ((cursor = find_prev( win->cursor )) == NULL)
         break;
      win->cursor = cursor;
   }
   show_curl_line( win );
   sync( win );
}


/*
 * Name:    goto_bottom
 * Purpose: To move the cursor to the bottom of the current window.
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 */
void goto_bottom( windows *window )
{
text_ptr cursor;
register windows *win;  /* put window pointer in a register */

   win = window;
   un_copy_line( win->cursor, win, TRUE );
   update_line( win );
   win->cursor = cpf( win->cursor );
   for (; win->cline < win->bottom_line; win->cline++,win->rline++) {
      if ((cursor = find_next( win->cursor )) == NULL)
         break;
      win->cursor = cursor;
   }
   show_curl_line( win );
   sync( win );
}


/*
 * Name:    set_tabstop
 * Purpose: To set the current interval between tab stops
 * Date:    October 1, 1989
 * Notes:   Tab interval must be reasonable, and this function will
 *           not allow tabs more than MAX_COLS / 2.
 */
void set_tabstop( windows *window )
{
char num_str[MAX_COLS];  /* tab interval as a character string */
int tab;                 /* new tab interval */
register int rc;

   rc = OK;
   while (rc == OK) {
      itoa( mode.tab_size, num_str, 10 );
      rc = get_name( "Tab interval: ", window->bottom_line, num_str,
                     g_display.message_color );
      if (rc == OK) {
         tab = atoi( num_str );
         if (tab < MAX_COLS/2) {
            mode.tab_size = tab;
            rc = ERROR;
         } else
            error( WARNING, window->bottom_line, "tab size too long" );
      }
   }
}


/*
 * Name:    show_line_col
 * Purpose: show current real line and column of current cursor position
 * Date:    June 5, 1991
 * Passed:  window:   information allowing access to the current window
 * Notes:   Blank old position and display new position.  current line and
 *          column may take up to 11 columns, which allows the display of
 *          999 columns and 9,999,999 lines.
 */
void show_line_col( windows *window )
{
int i, pline;
register int k;
char line_col[20], num[10];

   /*
    * blank out current line:column position.
    */
   memset( line_col, ' ', 12 );
   line_col[12] = '\0';

   /*
    * convert column to ascii and store in display buffer.
    */
   itoa( window->rcol+1, num, 10 );
   i = strlen( num ) - 1;
   for (k=11; i>=0; i--, k--)
      line_col[k] = num[i];

   /*
    * put in colon to separate line and column
    */
   line_col[k--] = ':';

   /*
    * convert line to ascii and store in display buffer.
    */
   ltoa( window->rline, num, 10 );
   i = strlen( num ) - 1;
   for (; i>=0; i--, k--)
      line_col[k] = num[i];

   /*
    * find line to start line:column display then output
    */
   pline = window->top_line - 1;
   s_output( line_col, pline, MAX_COLS-12, g_display.head_color );

   show_asterisk( window->file_info->modified, pline );
}


/*
 * Name:    show_asterisk
 * Purpose: give user an indication if file is dirty
 * Date:    September 16, 1991
 * Passed:  modified:   boolean - TRUE if modified, FALSE otherwise
 *          line:       line to display asterisk
 */
void show_asterisk( int modified, int line )
{
   c_output( modified ? '*' : ' ', 4, line, g_display.head_color );
}


/*
 * Name:    toggle_overwrite
 * Purpose: toggle overwrite-insert mode
 * Date:    September 16, 1991
 * Passed:  window
 */
void toggle_overwrite( windows *arg_filler )
{
   mode.insert = !mode.insert;
   show_insert_mode( );
   set_cursor_size( mode.insert ? g_display.insert_cursor :
                    g_display.overw_cursor );
}


/*
 * Name:    toggle_indent
 * Purpose: toggle indent mode
 * Date:    September 16, 1991
 * Passed:  window
 */
void toggle_indent( windows *arg_filler )
{
   mode.indent = !mode.indent;
   show_indent_mode( );
}


/*
 * Name:    set_left_margin
 * Purpose: set left margin for word wrap
 * Date:    November 27, 1991
 * Passed:  window
 */
void set_left_margin( windows *window )
{
register int rc;
char temp[80];

   itoa( mode.left_margin + 1, temp, 10 );
   rc = get_name("Enter new left margin (must be less than right margin) : ",
                   window->bottom_line, temp, g_display.message_color );
   if (rc == OK) {
      rc = atoi( temp ) - 1;
      if (rc < 0 || rc >= mode.right_margin)
         error( WARNING, window->bottom_line, "Left margin out of range." );
      else {
         mode.left_margin = rc;
         show_left_margin( );
      }
   }
}


/*
 * Name:    set_right_margin
 * Purpose: set right margin for word wrap
 * Date:    November 27, 1991
 * Passed:  window
 */
void set_right_margin( windows *window )
{
register int rc;
char temp[80];

   itoa( mode.right_margin + 1, temp, 10 );
   rc = get_name("Enter new right margin (must be greater than left margin) : ",
                   window->bottom_line, temp, g_display.message_color );
   if (rc == OK) {
      rc = atoi( temp ) - 1;
      if (rc <= mode.left_margin || rc > MAX_LINE_LENGTH)
         error( WARNING, window->bottom_line, "Right margin out of range." );
      else {
         mode.right_margin = rc;
         show_right_margin( );
      }
   }
}


/*
 * Name:    set_paragraph_margin
 * Purpose: set column to begin paragraph
 * Date:    November 27, 1991
 * Passed:  window
 * Notes:   paragraph may be indented, flush, or offset.
 */
void set_paragraph_margin( windows *window )
{
register int rc;
char temp[80];

   itoa( mode.parg_margin + 1, temp, 10 );
   rc = get_name("Enter paragraph margin (must be less than right margin) : ",
                   window->bottom_line, temp, g_display.message_color );
   if (rc == OK) {
      rc = atoi( temp ) - 1;
      if (rc < 0 || rc >= mode.right_margin)
         error( WARNING, window->bottom_line, "Paragraph margin out of range.");
      else {
         mode.parg_margin = rc;
         show_paragraph_margin( );
      }
   }
}


/*
 * Name:    toggle_crlf
 * Purpose: toggle crlf mode
 * Date:    November 27, 1991
 * Passed:  window
 */
void toggle_crlf( windows *arg_filler )
{
   mode.crlf = (mode.crlf == CRLF) ? LF : CRLF;
   show_crlf_mode( );
}


/*
 * Name:    toggle_ww
 * Purpose: toggle word wrap mode
 * Date:    November 27, 1991
 * Passed:  window
 */
void toggle_ww( windows *arg_filler )
{
   mode.word_wrap = !mode.word_wrap;
   show_wordwrap_mode( );
}


/*
 * Name:    toggle_trailing
 * Purpose: toggle eleminating trainling space at eol
 * Date:    November 25, 1991
 * Passed:  window
 */
void toggle_trailing( windows *arg_filler )
{
   mode.trailing = !mode.trailing;
   show_trailing( );
}


/*
 * Name:    toggle_z
 * Purpose: toggle writing control z at eof
 * Date:    November 25, 1991
 * Passed:  window
 */
void toggle_z( windows *arg_filler )
{
   mode.control_z = !mode.control_z;
   show_control_z( );
}


/*
 * Name:    toggle_eol
 * Purpose: toggle writing eol character at eol
 * Date:    November 25, 1991
 * Passed:  window
 */
void toggle_eol( windows *arg_filler )
{
register file_infos *file;

   mode.show_eol = !mode.show_eol;
   for (file=g_status.file_list; file != NULL; file=file->next)
      file->dirty = GLOBAL;
}


/*
 * Name:    toggle_search_case
 * Purpose: toggle search case
 * Date:    September 16, 1991
 * Passed:  window
 */
void toggle_search_case( windows *arg_filler )
{
   bm.search_case = (bm.search_case == IGNORE) ? MATCH : IGNORE;
   show_search_case( );
   if (bm.search_defined == OK)
      build_boyer_array( );
}


/*
 * Name:    toggle_sync
 * Purpose: toggle sync mode
 * Date:    January 15, 1992
 * Passed:  window
 */
void toggle_sync( windows *arg_filler )
{
   mode.sync = !mode.sync;
   show_sync_mode( );
}


/*
 * Name:    sync
 * Purpose: carry out cursor movements in all windows opened to the same file
 * Date:    January 15, 1992
 * Passed:  window
 * Notes:   turn off sync when we do this so we don't get into a recursive loop.
 *          All cursor movement commands un_copy_line before moving the
 *          cursor off the current line.   You MUST make certain that the
 *          current line is uncopied in the task routines before calling sync.
 */
void sync( windows *window )
{
register windows *wp;
register file_infos *fp;

   if (mode.sync && mode.sync_sem) {
      mode.sync_sem = FALSE;
      wp = g_status.window_list;
      while (wp != NULL) {
         if (wp->visible && wp != window) {
            (*do_it[g_status.command])( wp );
            show_line_col( wp );
         }
         wp = wp->next;
      }
      mode.sync_sem = TRUE;
      fp = g_status.file_list;
      while (fp != NULL) {
         if (fp->dirty != FALSE)
            fp->dirty = GLOBAL;
         fp = fp->next;
      }
   }
}


/*
 * Name:    editor
 * Purpose: Set up the editor structures and display changes as needed.
 * Date:    June 5, 1991
 * Notes:   Master editor routine.
 */
void editor( )
{
char *name;  /* name of file to start editing */
register windows *window;            /* current active window */
windows *above;             /* window above current */
windows *below;             /* window below current */
register file_infos *file;           /* temporary file structure */
int c;

   /*
    * Check that user specified file to edit, if not offer help
    */
   if (g_status.argc > 1)
      c = edit_next_file( g_status.current_window );
   else {
      name = g_status.rw_name;
      *name = '\0';
      c = get_name( "File name to edit : ", g_display.nlines, name,
                      g_display.text_color );
      if (c == ERROR || *name == '\0')
         return;
      if (c == OK) {
         c = edit_file( name );
         if (c == OK)
            c = initialize_window( );
      }
   }
   g_status.stop = c == OK ? FALSE : TRUE;
   if (c == OK) {
      window = g_status.current_window;
      redraw_screen( window );
      set_cursor_size( mode.insert ? g_display.insert_cursor :
                       g_display.overw_cursor );
   }

   /*
    * main loop - keep updating the display and processing any commands
    *  while user has not pressed the stop key
    */
   for (; g_status.stop != TRUE;) {
      window = g_status.current_window;

      /*
       * update all the other windows that point to file that has been changed
       */
      above = below = window;
      while (above->prev || below->next) {
         if (above->prev) {
            above = above->prev;
            if (above->visible) {
               file = above->file_info;
               if (file->dirty == GLOBAL || file->dirty == NOT_LOCAL) {
                  display_current_window( above );
                  show_size( above );
               }
               show_asterisk( file->modified, above->top_line-1 );
            }
         }
         if (below->next) {
            below = below->next;
            if (below->visible) {
               file = below->file_info;
               if (file->dirty == GLOBAL || file->dirty == NOT_LOCAL) {
                  display_current_window( below );
                  show_size( below );
               }
               show_asterisk( file->modified, below->top_line-1 );
            }
         }
      }
      file = window->file_info;
      if (file->dirty == LOCAL || file->dirty == GLOBAL)
         display_current_window( window );
      for (file=g_status.file_list; file != NULL; file=file->next)
         file->dirty = FALSE;

      /*
       * Set the cursor position at window->ccol, window->cline.  Show the
       * user where in the file the cursor is positioned.
       */
      xygoto( window->ccol, window->cline );
      show_line_col( window );

      /*
       * Get a key from the user.  Look up the function assigned to that key.
       * All regular text keys are assigned to function 0.  Text characters
       * are less than 0x100, decimal 256, which includes the ASCII and
       * extended ASCII character set.
       */
      g_status.key_pressed = getkey( );
      g_status.command = getfunc( g_status.key_pressed );
      if (g_status.wrapped) {
         g_status.wrapped = FALSE;
         show_search_message( CLR_SEARCH, g_display.mode_color );
      }
      if (g_status.command >= 0 && g_status.command < NUM_FUNCS)
         (*do_it[g_status.command])( window );
   }
   cls( );
   xygoto( 0, 0 );
}
